%MACRO QC(
dsn=, /* Input data set name */
use_lib=WORK,
/*** use_lib is used for several things.
	It's the library that holds the raw data. It's the library
	where the quality control (QC) files &use_lib..bad_ccs, &use_lib..missing_an_allele, and
	&use_lib..snp_warnings will be deposited. 
	See this section in the MAXTEST macro for more details about &use.lib ***/ 
FROM_MACRO=MAX_TEST,
/*** FROM_MACRO should be either MAX_TEST or PREPARE4STEPDOWN depending whether
DO_QC is called from MAX_TEST or from PREPARE4STEPDOWN, repectively.
/******************************************************************/
out_dsn=good_data
);
/********************************************************************/
/*** This MACRO does the following:
checks that current MACRO is being called from either MAX_TEST or PREPARE4STEPDOWN,

check that data set dsn is in library use_lib,

check that a1, a2, aff_stat, locus and ID are on input data set (also,
	determination of type, numeric or alpha, of a1 and a2),

sorts SNP data by locus and id, renames vars on original data set to be
	"CASE", "ID", "locus", "a1" and "a2". If A1 and A2 are character, makes sure that all chars
	are A, G, C, T or missing. If A1 and A2 are numeric, makes sure that all nums
	are 0, 1 or missing. In cases where subject has one missing allele value,
	lists that subject in ds "missing_an_allele".

makes sure that all SNPs on input data set are bi-allelic, edits out SNPs that aren't.
Edits out SNPs which have only cases or only controls.

if current MACRO is called from PREPARE4STEPDOWN, checks that locus names contain only numbers, 
	letters and underscores. As well, because of the way that we manipulate SNP names 
	in PREPARE4STEPDOWN, makes sure that no SNP name terminates in 'zjk2'. Also, adds
	SNP_count info to the file 'params'.
/********************************************************************/
/*** Start doing data quality checks ***/
%global _abort_;
%if &FROM_MACRO^=MAX_TEST and &FROM_MACRO^=PREPARE4STEPDOWN %then %do;
	%let _abort_=1;
	%put ;
	%put FROM_MACRO must be either "MAX_TEST" or "PREPARE4STEPDOWN";
	%put Currently, the value of FROM_MACRO is "&FROM_MACRO";
	%put ;
	%goto done;
%end;

%if &FROM_MACRO=MAX_TEST %then %let LOCUS_LENGTH_TXT=LEN;

%if &FROM_MACRO=PREPARE4STEPDOWN %then %let LOCUS_LENGTH_TXT=LEN+4;

/******************************************************************/

/** Check that data set dsn is in library use_lib ***/

/******************************************************************/
%let _abort_=1;
%if %sysfunc(exist(&use_lib..&dsn)) %then
      %let _abort_=0;

%if &_abort_=1 %then %do;
	%put;
	%put &FROM_MACRO.: "&DSN" is not in library "&use_lib".;
	%put;
	%goto DONE;
%end;
/******************************************************************/

/*	Check that a1, a2, aff_stat, locus and ID are on input data set. 
	Determine type (numeric or alpha) of a1 and a2. */

/******************************************************************/
ods output variables=tmp; 
proc contents data=&use_lib..&dsn ; run;
quit;
%GLOBAL LENGTH_LOCUS;
data tmp; set tmp END=LAST; 
retain found_a1 found_a2 found_aff_stat found_locus 0;
if variable="&a1" then do;
	found_a1=1;
	call symput('_type_a1_', trim(left(type)));
end;
if variable="&a2" then do;
	found_a2=1;
	call symput('_type_a2_', trim(left(type)));
end;
if variable="&aff_stat" then found_aff_stat=1;
if variable="&locus" then found_locus=1;
if variable="&IDENT" then found_ident=1;
if last then do;
	if found_a1=0 or found_a2=0 or found_aff_stat=0 or found_locus=0 or found_ident=0 then 
		call symput('_abort_', 1); 
end; 
run;

%if &_abort_=1 %then %do;
	%put;
	%put &FROM_MACRO.: One of the variables "&a1", "&a2", "&aff_stat" or "&locus";
	%put not found on data set "&dsn" in library "&use_lib".;
	%put;
	%goto DONE;
%end;

%if &_type_a1_ ^= &_type_a2_ %then %do;
	%let _abort_=1;
	%put;
	%put &FROM_MACRO.: The variables "&a1" and "&a2" are not of the same type.; 
	%put;
%end;

%if &_type_a1_ ^= Num and &_type_a1_ ^= Char %then %do;
	%let _abort_=1;
	%put;
	%put &FROM_MACRO.: The variables "&a1" and "&a2" are not of type "Char" or "Num".;
	%put;
%end;
	
%if &_abort_=1 %then %goto DONE;

/******************************************************************/

/*	Sort SNP data by locus and id. Rename vars on original data set to be
	"CASE", "ID", "locus", "a1" and "a2". If A1 and A2 are character, make sure that all chars
	are A, G, C, T or missing. If A1 and A2 are numeric, make sure that all nums
	are 0, 1 or missing. Remove records with missing values. */

/******************************************************************/
PROC SORT data=&use_lib..&dsn(KEEP=&aff_stat. &IDENT &locus &a1 &a2
	rename=(&aff_stat.=CASE &IDENT=ID &locus=locus &a1=a1 &a2=a2)) 
	out=c2; BY LOCUS ID; RUN;
/*** If A1 and A2 are char... ***/
%if "&_type_a1_"="Char" %then %do;
/*** In this case, initial data set must have all alleles coded as A,C,G ot T or 
left blank for missing. Start by editing out all missings. ***/
	data c2b; set c2; 
	if ((a1="G" or a1="T" or a1="C" or A1="A") and (a2="G" or a2="T" or a2="C" or A2="A"));
	run;
%end;

%else %do;
/*** If A1 and A2 are numeric... ***/
	data c2b; set c2; 
	if ((a1=0 or a1=1) and (a2=0 or a2=1));
	run;
%end;

/* Remember to print this out later */
data missing_an_allele; merge c2(in=A) c2b(in=B); by LOCUS ID; 
if A and ^B;
RUN;

/******************************************************************/

/*** Make sure that all SNPs on input data set are bi-allelic ***/
/*** Assumption is that data are in sas genetics 'tall' format ***/
/*** Note that 'maybe_good_SNPs' contains no info about particular subjects ***/
/*** Edit out SNPs that aren't bi-allelic. ***/
/*** Edit out SNPs which have only cases or only controls ***/
/******************************************************************/

data maybe_good_SNPs(keep=locus min_allele maj_allele ) SNP_WARNings(KEEP=locus WARNING); 
retain min_allele maj_allele BOTH_SET WARNING_FOUND;
/*** min_allele and maj_allele are chosen arbitrarily here. The names actually
become meaningful when tmp_max_data is created below. ***/
length warning $ 50;
set c2b(keep= locus id a1 a2 case);
by locus id;
if first.locus then do;
/*** Reset BOTH_SET and WARNING_FOUND. If a's are the same, initialize just min_allele. 
Otherwise, initialize both min_allele and maj_allele. ***/
	BOTH_SET=0; WARNING_FOUND=0;
	IF A1=A2 THEN min_allele=A1;
	ELSE DO; 
		min_allele=A1; maj_allele=A2;  BOTH_SET=1;
		OUTPUT maybe_good_SNPs;
	END;
END;
IF ^first.locus and WARNING_FOUND=0 and BOTH_SET=0 AND A1^=A2 THEN DO;
/*** if ^first.locus and WARNING_FOUND=0 and BOTH_SET=0 AND A1^=A2
then we're still looking for the 2nd allele, all previous alleles
for this SNP are the same, this subject furnishes the 2nd allele 
for this SNP ***/ 
	IF ^(min_allele=A1 OR min_allele=A2) THEN DO;
		WARNING='SNP ' || locus || 'has more than 2 alleles';
		WARNING_FOUND=1;
		output SNP_WARNings;
	end;
	else do;
		if min_allele=a1 then maj_allele=a2;
		else maj_allele=a1;
		BOTH_SET=1;
		output maybe_good_SNPs;
	end;
end;
IF ^first.locus and WARNING_FOUND=0 and BOTH_SET=1 THEN DO;
/*** if ^first.locus and WARNING_FOUND=0 and BOTH_SET=1 THEN just make sure
that both alleles for this subject correspond to those already seen ***/  
	if^((A1=min_allele or A1=maj_allele) and (A2=min_allele or A2=maj_allele)) then do;
		WARNING='SNP ' || locus || 'has more than 2 alleles';
		WARNING_FOUND=1;
		output SNP_WARNings;
	end;
end;
IF last.locus and WARNING_FOUND=0 and BOTH_SET=0 THEN DO;
/*** IF last.locus and WARNING_FOUND=0 and BOTH_SET=0 THEN this SNP is mon-allelic
and is of no further interest ***/
	WARNING='SNP ' || locus || 'is mono-allelic';
		WARNING_FOUND=1;
		output SNP_WARNings;
end;
run;
/*** Edit out those SNPs that aren't bi-allelic ***/
data good_SNPs(drop=warning);  
merge maybe_good_SNPs SNP_WARNings(in=A) end=last; 
by locus; if ^A; 
run;

/*** Now merge back good_SNPs, which contains info on SNPs but not
subjects, with the subject-specific data ***/
data OK_so_far; merge c2b good_SNPs(in=A); if A; by locus; run;

/*** Edit out SNPs which have only cases or only controls ***/
data bad_CCs(keep=locus); set OK_so_far; by locus; retain found_case found_control; 
if first.locus then do;
	found_case=0; found_control=0;
end;
if case=1 then found_case=1; 
if case=0 then found_control=1; 
if last.locus and (found_case=0 or found_control=0) then output;
run;

/* Create data set which is now QC'ed as well as data set which can be used to 
decode locus using SNP_count */
/***********************************************************************/

/* Also, get the number distinct SNPs, 'SNP_count' */
%global SNP_count;
/***********************************************************************/
data tmp; merge OK_so_far bad_CCs(in=a); by locus; if ^a; run;

data &out_dsn(drop=locus rename=(SNP_count=locus_no)) 
	&use_lib..locus_key(keep=locus SNP_count rename=(SNP_count=locus_no))
	&use_lib..params(keep=SNP_count); 
set tmp end=last ; by locus;
retain SNP_count 0;
if first.locus then SNP_count=SNP_count+1;
output &out_dsn;
if last.locus then output &use_lib..locus_key;
if last then do;
	call symput('SNP_count', SNP_count);
	output &use_lib..params;
end;
run;


/****************************************************************************************/

/*** End of data quality checks ***/

/****************************************************************************************/

/*** Clean up data sets ***/
data &use_lib..bad_ccs; set bad_ccs; run;
data &use_lib..missing_an_allele; set missing_an_allele; run;
data &use_lib..snp_warnings; set snp_warnings; run;

proc datasets; 
delete c2 c2b maybe_good_snps good_snps OK_so_far
	bad_ccs missing_an_allele snp_warnings tmp;
run; quit;

%put ****************************************************************************************;
%put;
%put &FROM_MACRO.: Initial checks look good.;
%put Remember to look at the data sets SNP_WARNings, bad_ccs, and ;
%put missing_an_allele which are all deposited in library &use_lib.;
%put;
%put ****************************************************************************************;
%done:

%mend; /*** End of macro DO_QC ***/






